#################################################################################
#####                                                                       #####
##### This is the code for Step 2: regression to estimate the non-pure rows #####
#####                                                                       #####
#################################################################################
library(linprog)

EstC <- function(Sigma, AI, diagonal) {
  # Estimate C. If diagonal=True, estimate only diagonal elements.
  #
  # Args: 
  #   Sigma: p by p covariance matrix.
  #   AI: p by K matrix.
  #   diagonal: TRUE or FALSE.
  #
  # Returns: 
  #   K by K estimated C_hat 
  K <- ncol(AI)
  C <- diag(0, K, K)
  for (i in 1:K) {
    groupi <- which(AI[ ,i] != 0)
    if (length(groupi) == 1)
      stop("Singleton pure node is not allowed...Need to reselect delta...\n")
    sigmai <- as.matrix(abs(Sigma[groupi,groupi]))
    tmpEntry <- sum(sigmai) - sum(diag(sigmai))
    if (length(groupi) == 1) {
      C[i,i] <- tmpEntry
    } else {
      C[i,i] <- tmpEntry / (length(groupi) * (length(groupi) - 1))
    }
    if (!diagonal && i<K) {
      for (j in (i+1):K) {
        groupj <- which(AI[ ,j]!=0)
        # adjust the sign for each row
        sigmaij <- AI[groupi,i] * as.matrix(Sigma[groupi,groupj]) 
        sigmaij <- t(AI[groupj,j] * t(sigmaij))
        tmpEntry <- sum(sigmaij)
        C[i,j] <- C[j,i] <- tmpEntry / (length(groupi) * length(groupj))
      }
    }
  }
  return(C)
}


EstY <- function(Sigma, AI, pureVec) {
  # Estimate the Sigma_{TJ}_hat
  # 
  # Args: 
  #   Sigma: p by p matrix.
  #   AI: p by K matrix.
  #   pureVec: the vector of pure node indices.
  #
  # Returns: 
  #   Sigma_{TJ}_hat: K by |J| matrix.
  SigmaS <- AdjustSign(Sigma, AI)
  SigmaJ <- matrix(SigmaS[ ,-pureVec], nrow <- nrow(Sigma))
  SigmaTJ <- matrix(0, ncol(AI), nrow(AI) - length(pureVec))
  for (i in 1:ncol(AI)) {
    groupi <- which(AI[ ,i] != 0)
    SigmaiJ <- as.matrix(SigmaJ[groupi, ])
    SigmaTJ[i, ] <- apply(SigmaiJ, 2, mean) # Average columns along the rows.
  }
  return(SigmaTJ)
}


AdjustSign <- function(Sigma, AI) {
  # Sign operation on matrix {@code Sigma} according to the sign of {@code AI}
  #
  # Args: 
  #   Sigma: p by p matrix; AI: p by K matrix
  #
  # Returns: 
  #   p by p matrix
  SigmaS <- matrix(0, nrow(AI), nrow(AI))
  for (i in 1:nrow(AI)) {
    index <- which(AI[i, ] != 0)
    if (length(index) != 0)
      SigmaS[i, ] = sign(AI[i,index]) * Sigma[i, ]
  }
  return(SigmaS)
}

EstAJInv <- function(Omega,Y,lbd) {
  # This function estimates the |J| by K matrix A_J by using soft thresholding.
  #
  # Args: 
  #   Omega: K by K estimated C^{-1}.
  #   Y: K by |J| reponse matrix.
  #   lbd: the chosen constant of the RHS constraint for soft-thresholding.
  #
  # Returns: 
  #   |J| by K matrix A_J.
  AJ <- matrix(0, ncol(Y), nrow(Y))
  for (i in 1:ncol(Y)) {
    Atilde <- Omega %*% as.matrix(Y[ ,i])
    AJ[i, ] <- LP(Atilde, lbd)
    if (sum(abs(AJ[i, ])) > 1)
      AJ[i,] <- AJ[i,] / sum(abs(AJ[i, ]))
  }
  return(AJ)
}


