####### This script contains functions to estimate the pure node set ##########
source("Utilities.R")

FindRowMax <- function(Sigma) {
  # Calculate the maximal absolute value for each row of the given matrix.
  #
  # Args:
  #   Sigma: p by p matrix
  # 
  # Returns:
  #   length p vector
  M <- rep(0, nrow(Sigma))
  for (i in 1:nrow(Sigma)) {
    Mi <- max(abs(Sigma[i, -i]))   # remove the diagonal element
    M[i] <- Mi
  }
  return(M)
}

FindRowMaxInd <- function(i, M, vector, delta) {
  # Calculate indices of ith row such that the absolute values of these indices
  # are within 2*delta from the maximal absolute value {@code M} of this row.
  # 
  # Args:
  #   i: integer denoting for the ith row.
  #   M: the maximal absolute value of the ith row.
  #   vector: the entries of this row.
  #   delta: numerical constant.
  #
  # Returns:
  #   a vector of indices.
  lbd <- 2*delta
  indices <- which(M - abs(vector) <= lbd)
  return(indices[which(indices != i)])
}

TestPure <- function(Sigma, rowInd, Si, Ms, delta) {
  # For given row, check if it is a pure node by iteratively checking the nodes
  # in Si. Return TRUE if the given row corresponds to a pure variable.
  #
  # Args:
  #   Sigma: p by p matrix.
  #   rowInd: integer index.
  #   Si: vector of indices.
  #   Ms: vector of largest absolute values of each rows in Si
  #   delta: numerical constant.
  # Returns:
  #   TRUE or FALSE.
  for (i in 1:length(Si)) {
    j <- Si[i]
    if (abs(abs(Sigma[rowInd, j]) - Ms[j]) > 2 * delta)
      return(FALSE)
  }
  return(TRUE)
}

UnionGroup <- function(groupList, Union) {
  # Union all the groups which have common variables
  newGroup <- list(groupList[[1]])
  for (i in 2:length(groupList)) {
    groupi <- groupList[[i]]
    inGroupFlag <- FALSE
    for (j in 1:length(newGroup)) {
      groupj <- newGroup[[j]]
      if (length(intersect(groupi, groupj)) != 0) {
        inGroupFlag <- TRUE
        if (Union) {
          newGroup[[j]] <- union(groupi, groupj)
        } else {
          newGroup[[j]] <- intersect(groupi, groupj)
        }
        break
      }
    }
    if (!inGroupFlag)
      newGroup <- append(newGroup, list(groupi))
  }
  return(newGroup)
}

CheckDisjoint <- function(groupList) {
  # Check if all the sets are disjoint
  if (length(groupList) == 1)
    return(TRUE)
  for (i in 1:(length(groupList) - 1)) {
    groupi <- groupList[[i]]
    for (j in (i+1):length(groupList)) {
      groupj <- groupList[[j]]
      if (length(intersect(groupi, groupj)) != 0)
        return(FALSE)
    }
  }
  return(TRUE)
}

FindPureNode = function(Sigma, delta, Ms, Union) {
  # Estimate list of pure node indices for given {@code Sigma} and {@delta}.
  # 
  # Args: 
  #   Sigma: p by p sample covariance matrix.
  #   delta: numerical constant.
  #   Ms: the largest absolute values of each row of Sigma.
  #   Union: flag for using union or intersection for p sets of pure nodes.
  #
  # Returns: 
  #   a list including:
  #     the list of the estimated pure node indices
  #     the vector of the estimated pure node indices
  G <- list()
  vec <- c()
  # Calculate the Si for those rows which pass the checking
  for (i in 1:nrow(Sigma)) {
    Si <- FindRowMaxInd(i, Ms[i], Sigma[i,], delta)
    if (length(Si) == 0)
      next
    pureFlag <- TestPure(Sigma, i, Si, Ms, delta) 
    if (pureFlag) {
      Gi <- c(Si, i)
      G <- append(G, list(Gi))
      names(G)[length(G)] <- i
    }
  }
  while(!CheckDisjoint(G))
    G <- UnionGroup(G, Union)
  return(list(pureInd = G, pureVec = unlist(G)))
}

FindSignPureNode <- function(pureList, Sigma) {
  # Estimate the sign subpartition of pure node sets. If there is an element 
  # of a list is empty, then a empty list will be put in that position
  # 
  # Args: 
  #   pureList: list of pure node indices (Example: list(c(1,2,3),c(4,5,6,7)))
  #   Sigma: p by p sample covariance
  #
  # Returns: 
  #   list of sign subpartition of pure node indices.
  #   Example: list(list(c(1,2),3),list(c(4,7),c(5,6)))
  signPureList <- list()
  for (i in 1:length(pureList))
  {
    purei <- sort(pureList[[i]])   ### For simulation purpose. Not necessary! 
    if (length(purei) == 1) {
      signPureList[[i]] <- list(pos = purei[1], neg = list())
      next
    }
    firstPure <- purei[1]
    pos <- firstPure
    neg <- c()
    for (j in 2:length(purei)) {
      purej <- purei[j]
      if (Sigma[firstPure, purej] < 0)
        neg <- c(neg, purej)
      else
        pos <- c(pos, purej)
    }
    if (length(pos) == 0) {
      pos <- list()
    } else {
      pos <- pos
    }
    if (length(neg) == 0) {
      neg <- list()
    } else {
      neg <- neg
    }
    signPureList[[i]] <- list(pos = pos, neg = neg)
  }
  return(signPureList)
}

RecoverAI <- function(estGroupList, p) {
  # Recover the estimated submatrix A_I by given the pure node group.
  # 
  # Args:
  #   estGroupList: list of group indices of the pure node with sign.
  #
  # Returns: 
  #   p by K matrix.
  K <- length(estGroupList)
  A <- matrix(0, p, K)
  for (i in 1:K) {
    groupi <- estGroupList[[i]]
    groupi1 <- groupi[[1]]       # group with positive sign
    for (j in groupi1)
      A[j,i] <- 1
    groupi2 <- groupi[[2]]
    if (length(groupi2) != 0) {
      for (k in groupi2) {
        A[k,i] <- -1
      }
    }
  }
  return(A)
}

EstAI <- function(Sigma,delta,Union) {
  # Use the given {@code delta} to calculate the fitted AI, pure variable 
  # indices in list form and vector form. Also return estimated Y and C for 
  # the following Dantzig estimation.
  #
  # Args: 
  #   Sigma: p by p covariance matrix.
  #   delta: delta to be used.
  #
  # Return: 
  #   the list including:
  #     AI: the p by K estimated AI
  #     pureVec: vector of the indices of estimated pure variables
  #     pureSignInd: list of the indices of estimated pure variables   
  Ms <- FindRowMax(Sigma)
  resultPure <- FindPureNode(Sigma, delta, Ms, Union)
  estPureIndices <- resultPure$pureInd
  estPureVec <- resultPure$pureVec
  estSignPureIndices <- FindSignPureNode(estPureIndices, Sigma)
  AI <- RecoverAI(estSignPureIndices, nrow(Sigma))
  return(list(AI = AI, pureVec = estPureVec, pureSignInd = estSignPureIndices))
}

