########################################################################################
#############              LOVE --- Overlapping Clustering               ############### 
########################################################################################

LOVE <- function(X, delta = seq(1.5, 2.5 ,0.1), HT = F, lbd = 1, mu = 1, 
                 diagonal = FALSE, stand = TRUE, N = 50, Union = TRUE, 
                 verbose = TRUE) {
  # Non-overlapping/Overlappig variable clustering for given data X where each 
  # row denotes a data point and columns for variables.
  #
  # Args: 
  #   X: n by p data matrix.
  #   delta: grid of leading constants for tuning delta. This grid needs to be
  #     multiplied by sqrt(log(p)/n).
  #   HT: If TRUE, then hard-thresholding is used for estimating A; 
  #     else soft-thresholding is used
  #   mu: the grid of leading constants for tuning the thresholding level
  #   diagonal: TRUE if the cov(Z) is diagonal; else FALSE. If TRUE, 
  #     OLS is used for solving A
  #   stand: True if centering is needed
  #   N: the number of cross validation used for delta
  #   Union: True if using "or" rule for estimating pure nodes; False if 
  #     using "and" rule
  #   verbose: if the process needs to be printed out.
  # 
  # Returns: 
  #   a list including: 
  #   K: the number of clusters.
  #   A: the estimated p by K matrix.
  #   group: the estimated group structure(indices for each cluster).
  #   pure: the vector of estimated pure variables.
  if (stand)  
    X <- scale(X, TRUE, FALSE)
  n <- nrow(X)
  p <- ncol(X)
  deltaGrids <- delta * sqrt(log(p) / n)
  if (length(deltaGrids) > 1) {
    if (verbose)
      cat("Selecting optimal delta by using data splitting...\n")
    optDelta <- median(replicate(N, CV_Delta(X, deltaGrids, Union, diagonal)))
  }else
    optDelta <- deltaGrids
  if (verbose) {
    cat("Finishing selecting optimal delta =", optDelta, "...\n")
    cat("Estimating pure rows...\n")
  }
  Sigma <- t(X) %*% X / nrow(X)
  resultAI <- EstAI(Sigma, optDelta, Union)
  AI <- resultAI$AI 
  if (length(resultAI$pureVec) == nrow(Sigma)) {
    group <- resultAI$pureSignInd
  } else {
    if (verbose)
      cat("Estimating C and Sigma_IJ...\n")
    C_hat <- EstC(Sigma, AI, diagonal) 
    Y <- EstY(Sigma, AI, resultAI$pureVec)
    if (diagonal) {
      if (verbose)
        cat("Estimating non-pure rows by using Ordinary Least Square...\n")
      C_inv <- diag(diag(C_hat)^{-1})
      AJ <- threshA(t(C_inv %*% Y), mu)
    } else {
      if (verbose)
        cat("Selecting the tuning parameter for estimating Omega...\n")
      optLbd <- lbd * optDelta
      if (length(lbd) > 1)
        optLbd <- median(replicate(50,CV_lbd(X, lbd * optDelta, AI, 
                                             resultAI$pureVec, diagonal)))
      if (verbose) {
        cat("Selecting the optimal lambda =", optLbd, "...\n")
        cat("Estimating Omega...\n")
      }
      Omega <- solveOmega(optLbd, C_hat)
      if (verbose && HT) {
        cat("Estimating non-pure rows by Hard Thresholding...\n")
      } else if (verbose) { 
        cat("Estimating non-pure rows by Soft Thresholding...\n")
      }
      if (HT) {
        AJ = threshA(t(Omega %*% Y), mu * optDelta * norm(Omega, "I"))
      } else {
        AJ = EstAJInv(Omega, Y, mu * optDelta * norm(Omega, "I"))
      }
    }
    AI[-resultAI$pureVec, ] <- AJ  
    group <-recoverGroup(AI) 
  }
  return(list(K = ncol(AI), A = AI, group = group, pure = resultAI$pureVec))
}