################################################################################
######                                                                   #######
######   These are the code which convert the original problem into LP   #######
######                                                                   #######
################################################################################
####   Given a symmetric p by p matrix Omega, now the parameters are the upper-
####   triangle elements.The vector of parameters has length equal to p*(p+1)/2.
####   eg:  Omega = (c11, c12, c13)
####                (c21, c22, c23)
####                (c31, c32, c33)
####   The vector of pars we want to estimate is: 
####   par_vec = c(c11(+),c11(-),c12(+),c12(-),c13(+),c13(-),
####               c22(+),c22(-),c23(+),c23(-),c33(+),c33(-)).
####   The following three functions are aimed to find the linear constraint 
####   matrix Amat such that Amat*bvec is the left hand side. 

setConstr <- function(i, vec) {
  # For a given ith row of the matrix Omega and {@code vec}, find the vec_constr
  # such that t(vec_constr) %*% par_vec = 
  #           t(vec) %*% (ci1(+),ci1(-),ci2(+),ci2(-),ci3(+),ci3(-)).
  #
  # Args: 
  #   i: is an index of each row, 1 <= i <= n.
  #   vec: length 2*p vector containing the values to be modified. 
  #
  # Returns: 
  #   length p*(p+1) vector. 
  #   eg: |C21| + |C22| + |C23| = t(c(1,1,1,1,1,1) %*% c(c21(+),c21(-),c22(+),c22(-),c23(+),c23(-)) 
  #                           = t(c(0,0,1,-1,0,0,1,-1,1,-1,0,0)) %*% par_vec 
  K <- length(vec) / 2
  constr <- c()
  for (j in 1:K) {
    if (j <= i - 1) {
      tmpConstr <- rep(0, 2 * (K - j + 1))
      tmpInd <- i - j + 1
      tmpConstr[(2 * tmpInd - 1):(2 * tmpInd)] = vec[(2 * j - 1):(2 * j)]
      constr <- c(constr, tmpConstr)
    } else {
      tmpConstr <- vec[(2 * j - 1):(2 * K)]
      constr <- c(constr, tmpConstr)
      break
    }
  }
  constr <- c(constr, rep(0, K * (K + 1) - length(constr)))
  return(constr)
}

lInfConstr <- function(p) {
  # Get the Amat constraint matrix of Inf-1 norm, i.e. the l_1 norm of each row.
  #
  # Args: 
  #   p: numerical constant denoting the dimension of Omega.
  #
  # Returns: 
  #   p by p(p+1) matrix.
  constrMat <- matrix(0, p, p * (p + 1))
  vec <- rep(1, 2 * p)
  for (i in 1:p) {
    constrMat[i, ] = setConstr(i, vec)
  }
  return(constrMat)
}

prodConstr <- function(C) {
  # Get the Amat constraint matrix for the product Omega %*% C based on 
  # diagonal and off-diagonal such that:
  #       t(A_i) %*% par_vec <- t(Omega_i) %*% C_i, 
  # for i belonging to the diagonal set. Similar for the off-diagonal, 
  # the order for each row of off-diagonal set is from the first row to 
  # the last row and ordered by row.
  # 
  # Args:
  #   C: K by K matrix.
  # Return:
  #   a list including:
  #     the constraint matrices of diagonal and offdiagonal.
  K <- nrow(C)
  constrDiag <- matrix(0, K, K * (K + 1))
  constrOffDiag <- matrix(0, K * (K - 1) / 2, K * (K + 1))
  counter <- 0
  for (i in 1:K) {
    for (j in i:K) {
      vec <- rep(C[ ,j], each = 2) * c(1,-1)
      if (j == i) {
        constrDiag[j, ] <- setConstr(j, vec)
      } else {
        counter <- counter + 1
        constrOffDiag[counter, ] <- setConstr(i, vec)
      }
    }
  }
  return(list(diag = constrDiag, off = constrOffDiag))
}

genConstr <- function(lbd, C) {
  # Obtain the Amat and bvec for a given C and lbd.
  K <- nrow(C)
  A <- diag(-1, nrow = 1 + K * (K + 1))
  constr1 <- cbind(-1, lInfConstr(K))
  A <- rbind(A, constr1)
  b <- c(rep(0, K * (K + 1) + 1), rep(0, nrow(constr1)))
  result <- prodConstr(C)
  constr2 <- rbind(cbind(-lbd, result$diag), cbind(-lbd, -result$diag))
  b <- c(b, rep(c(1,-1), each = K))
  constr3 <- rbind(cbind(-lbd, result$off), cbind(-lbd, -result$off))
  b <- c(b, rep(0, 2 * nrow(result$off)))
  A <- rbind(A, constr2, constr3)
  return(list(A = A, b = b))
}

solveOmega <- function(lbd, C) {
  # For a given lbd and C, solve the C^{-1}. 
  # Require: C should be symmetric and square. 
  cvec <- c(1, rep(0, nrow(C) * (nrow(C) + 1)))
  constrs <- genConstr(lbd, C)
  lpResult <- solveLP(cvec, constrs$b, constrs$A, lpSolve = T)$solution[-1]
  while (length(lpResult) == 0) {
    cat("The penalty lambda =", lbd, "is too small and increased by 0.01...\n")
    lbd <- lbd + 0.01
    constrs <- genConstr(lbd, C)
    lpResult <- solveLP(cvec, constrs$b, constrs$A, lpSolve = T)$solution[-1]
  }
  omegas <- rep(0, nrow(C) * (nrow(C) + 1) / 2)
  for (i in 1:length(omegas)) {
    omegas[i] <- lpResult[(2 * i - 1)] - lpResult[2 * i]
  }
  Omega <- matrix(0, nrow(C), nrow(C))
  Omega[lower.tri(Omega,diag = T)] <- omegas
  Omega <- t(Omega)
  Omega[lower.tri(Omega)] <- t(Omega)[lower.tri(Omega)]
  return(Omega)
}


LP <- function(y, lbd) {
  # Use LP to solve problem:
  #   beta^+ - beta^- \leq lbd + y
  #   - beta^+ + beta^- \leq lbd - y
  #   beta^+ \ge 0; beta^- \ge 0
  #
  # Args: 
  #   y: K by 1 vector. 
  #   lbd: positive contant.
  #
  # Returns:
  #   K by 1 vector
  K <- length(y)
  cvec <- rep(1, 2 * K)
  bvec <- c(lbd + y, lbd - y, rep(0, 2 * K))
  C <- matrix(0, K, 2 * K) 
  for (i in 1:K) {
    indices <- c(i,i + K)
    C[i,indices] = c(1,-1)
  }
  Amat <- rbind(C, -C, diag(-1, nrow=2 * K))
  LPsol <- solveLP(cvec, bvec, Amat, lpSolve = T)$solution
  beta <- LPsol[1:K] - LPsol[(K + 1):(2 * K)]
  return(beta)
}
