##### Parameter tuning and Cross validation ######
source("Utilities.R")

CV_Delta <- function(X, deltaGrids, Union, diagonal) {
  # Cross validation for choosing delta. For each delta from the given grids, 
  # first split the data into two datasets. Obtain I, AI and C. Then calculate 
  # AICAI^T and choose delta which minimizes the criterion 
  #                   AICAI^t - Sigma(dataset 2)
  #
  # Args: 
  #   X: n by p matrix.
  #   deltaGrids: vector of numerical constants.
  #   Union: TRUE or FALSE for estimating the pure variables.
  #     (For details, see doc of FindPureNode.)
  #   diagonal: TRUE or FALSE for the diagonal structure of C.
  # 
  # Returns: 
  #   the selected optimal delta
  sampInd <- sample(dim(X)[1], floor(dim(X)[1] / 2))
  X1 <- X[sampInd, ]
  X2 <- X[-sampInd, ]
  Sigma1 <- t(X1) %*% X1 / nrow(X1)
  Sigma2 <- t(X2) %*% X2 / nrow(X2)
  Ms <- FindRowMax(Sigma1)
  loss <- c()
  for (i in 1:length(deltaGrids)) {
    resultFitted <- CalFittedSigma(Sigma1, deltaGrids[i], Ms, Union, diagonal)
    fittedValue <- resultFitted$fitted   
    estPureVec <- resultFitted$pureVec
    if (is.null(dim(fittedValue)) && fittedValue == -1) {
      loss[i] <- Inf
    } else {
      subSigma2 <- Sigma2[estPureVec, estPureVec]
      denom <- length(estPureVec) * (length(estPureVec)-1)
      loss[i] <- 2 * offSum(subSigma2, fittedValue) / denom
    }
  } 
  return(deltaGrids[median(which(loss == min(loss)))])
}


CalFittedSigma <- function(Sigma, delta, Ms, Union, diagonal) {
  # Calculate the fitted value of A_ICA_I^T for given Sigma and delta.
  # 
  # Args: 
  #   Sigma: p by p covariance matrix. 
  #   delta: given parameter. 
  #   Ms: the calculated maximal values of Sigma per row.
  #   Union: TRUE or FALSE.
  #   diagonal: TRUE or FALSE.
  #
  # Returns: 
  #   a list including:
  #     pureVec: vector of the indices of estimated pure variables.
  #     fitted: fitted value of A_ICA_I^T 
  resultPureNode <- FindPureNode(Sigma, delta, Ms, Union)
  estPureIndices <- resultPureNode$pureInd
  if (singleton(estPureIndices))
    return(list(pureVec = NULL, fitted = -1))
  estSignPureIndices <- FindSignPureNode(estPureIndices, Sigma)
  AI <- RecoverAI(estSignPureIndices, nrow(Sigma))
  C <- EstC(Sigma, AI, diagonal)
  if (length(estPureIndices) == 1) {
    fitted <- -1
  } else {
    subAI <- AI[resultPureNode$pureVec, ]
    fitted <- subAI %*% C %*% t(subAI)
  }
  return(list(pureVec = resultPureNode$pureVec, fitted = fitted))
}


CV_lbd <- function(X, lbdGrids, AI, pureVec, diagonal) {
  # Use Cross-validation to select lambda for estimating Omega. Split the data
  # into two parts. Estimating C on two datasets. Then, for each lambda, calcu-
  # late Omega on the first dataset and calculate the loss on the second dataset.
  # Find the lambda which gives the smallest loss of 
  #                 <C,Omega> - log(det(Omega))
  #
  # Args: 
  #   X: n by p matrix.
  #   lbdGrids: vector of numerical constants.
  #   AI: p by K matrix.
  #   pureVec: vector of indices of pure variables.
  #   diagonal: TRUE or FALSE for the diagonal structure of C.
  # 
  # Returns: 
  #   the selected optimal lambda
  sampInd <- sample(nrow(X), floor(nrow(X) / 2))
  X1 <- X[sampInd, ]
  X2 <- X[-sampInd, ]
  Sigma1 <- t(X1) %*% X1 / nrow(X1)
  Sigma2 <- t(X2) %*% X2 / nrow(X2)
  C1 <- EstC(Sigma1, AI, diagonal)
  C2 <- EstC(Sigma2, AI, diagonal)
  loss <- c()
  for (i in 1:length(lbdGrids)) {
    Omega <- solveOmega(lbdGrids[i], C1)
    det_Omega <- det(Omega)
    if (det_Omega <= 0) {
      loss[i] <- Inf
    } else {
      loss[i] <- sum(Omega * C2) - log(det_Omega)
    }
  }
  return(lbdGrids[which.min(loss)])
}
